@@ -31,7 +31,7 @@ and source-highlight.
 
 Please note, that we're not aiming for complete docbook-xml support. All tags
 used in the generated xml are of course handled. More tags used in handwritten
-xml can be easilly supported, but for some combinations of tags we prefer
+xml can be easily supported, but for some combinations of tags we prefer
 simplicity.
 
 TODO:
@@ -84,7 +84,10 @@ sudo pip3 install anytree lxml pygments
 
 Example invocation:
 cd tests/bugs/docs/
-../../../gtkdoc-mkhtml2 tester tester-docs.xml
+mkdir -p db2html
+cd dbhtml
+../../../../gtkdoc-mkhtml2 tester ../tester-docs.xml
+cd ..
 xdg-open db2html/index.html
 meld html db2html
 
@@ -93,8 +96,6 @@ cd tests/bugs/docs/;
 rm html-build.stamp; time make html-build.stamp
 """
 
-import argparse
-import errno
 import logging
 import os
 import shutil
@@ -104,19 +105,9 @@ from anytree import Node, PreOrderIter
 from copy import deepcopy
 from glob import glob
 from lxml import etree
-from pygments import highlight
-from pygments.lexers import CLexer
-from pygments.formatters import HtmlFormatter
 from timeit import default_timer as timer
 
-from . import config, fixxref
-
-# pygments setup
-# lazily constructed lexer cache
-LEXERS = {
-    'c': CLexer()
-}
-HTML_FORMATTER = HtmlFormatter(nowrap=True)
+from . import config, highlight, fixxref
 
 
 class ChunkParams(object):
@@ -135,7 +126,7 @@ DONT_CHUNK = float('inf')
 # docbook xsl does.
 #
 # For toc levels see http://www.sagehill.net/docbookxsl/TOCcontrol.html
-# TODO: this list has also a flag that controls wheter we add the
+# TODO: this list has also a flag that controls whether we add the
 # 'Table of Contents' heading in convert_chunk_with_toc()
 CHUNK_PARAMS = {
     'appendix': ChunkParams('app', 'book'),
@@ -221,13 +212,14 @@ def gen_chunk_name(node, chunk_params):
     return name
 
 
-def get_chunk_titles(module, node):
+def get_chunk_titles(module, node, tree_node):
     tag = node.tag
     (title, subtitle) = TITLE_XPATHS.get(tag, TITLE_XPATHS['_'])
 
     ctx = {
         'module': module,
         'files': [],
+        'node': tree_node,
     }
     result = {
         'title': None,
@@ -272,7 +264,7 @@ def chunk(xml_node, module, depth=0, idx=0, parent=None):
     tag = xml_node.tag
     chunk_params = CHUNK_PARAMS.get(tag)
     if chunk_params:
-        title_args = get_chunk_titles(module, xml_node)
+        title_args = get_chunk_titles(module, xml_node, parent)
         chunk_name = gen_chunk_name(xml_node, chunk_params)
 
         # check idx to handle 'sect1'/'section' special casing and title-only
@@ -873,12 +865,8 @@ def convert_programlisting(ctx, xml):
     if xml.attrib.get('role', '') == 'example':
         if xml.text:
             lang = xml.attrib.get('language', ctx['src-lang']).lower()
-            if lang not in LEXERS:
-                LEXERS[lang] = get_lexer_by_name(lang)
-            lexer = LEXERS.get(lang, None)
-            if lexer:
-                highlighted = highlight(xml.text, lexer, HTML_FORMATTER)
-
+            highlighted = highlight.highlight_code(xml.text, lang)
+            if highlighted:
                 # we do own line-numbering
                 line_count = highlighted.count('\n')
                 source_lines = '\n'.join([str(i) for i in range(1, line_count + 1)])
@@ -1501,7 +1489,10 @@ def convert_glossary(ctx):
 def convert_index(ctx):
     node = ctx['node']
     # Get all indexdivs under indexdiv
-    indexdivs = node.xml.find('indexdiv').findall('indexdiv')
+    indexdivs = []
+    indexdiv = node.xml.find('indexdiv')
+    if indexdiv is not None:
+        indexdivs = indexdiv.findall('indexdiv')
 
     result = [
         HTML_HEADER % (node.title + ": " + node.root.title, generate_head_links(ctx)),
@@ -1627,6 +1618,22 @@ def generate_nav_nodes(files, node):
     return nav
 
 
+def convert_content(module, files, node, src_lang):
+    converter = convert_chunks.get(node.name)
+    if converter is None:
+        logging.warning('Add chunk converter for "%s"', node.name)
+        return []
+
+    ctx = {
+        'module': module,
+        'files': files,
+        'node': node,
+        'src-lang': src_lang,
+    }
+    ctx.update(generate_nav_nodes(files, node))
+    return converter(ctx)
+
+
 def convert(out_dir, module, files, node, src_lang):
     """Convert the docbook chunks to a html file.
 
@@ -1639,20 +1646,8 @@ def convert(out_dir, module, files, node, src_lang):
     logging.info('Writing: %s', node.filename)
     with open(os.path.join(out_dir, node.filename), 'wt',
               newline='\n', encoding='utf-8') as html:
-        ctx = {
-            'module': module,
-            'files': files,
-            'node': node,
-            'src-lang': src_lang,
-        }
-        ctx.update(generate_nav_nodes(files, node))
-
-        converter = convert_chunks.get(node.name)
-        if converter is not None:
-            for line in converter(ctx):
-                html.write(line)
-        else:
-            logging.warning('Add chunk converter for "%s"', node.name)
+        for line in convert_content(module, files, node, src_lang):
+            html.write(line)
 
 
 def create_devhelp2_toc(node):
@@ -1697,83 +1692,90 @@ def create_devhelp2_refsect3_keyword(node, base_link, title, name):
         create_devhelp2_condition_attribs(node))
 
 
+def create_devhelp2_content(module, xml, files):
+    title = ''
+    online_attr = ''
+    bookinfo_nodes = xml.xpath('/book/bookinfo')
+    if len(bookinfo_nodes):
+        bookinfo = bookinfo_nodes[0]
+        title = bookinfo.xpath('./title/text()')[0]
+        online_url = bookinfo.xpath('./releaseinfo/ulink[@role="online-location"]/@url')
+        if len(online_url):
+            online_attr = ' online="' + online_url[0] + '"'
+        # TODO: support author too (see devhelp2.xsl), it is hardly used though
+        # locate *.devhelp2 | xargs grep -Hn --color ' author="[^"]'
+    # TODO: fixxref uses '--src-lang' to set the language, we have this in options too
+    result = [
+        """<?xml version="1.0" encoding="utf-8" standalone="no"?>
+<book xmlns="http://www.devhelp.net/book" title="%s" link="index.html" author="" name="%s" version="2" language="c"%s>
+  <chapters>
+""" % (title, module, online_attr)
+    ]
+    # toc
+    result.extend(create_devhelp2_toc(files[0].root))
+    result.append("""  </chapters>
+  <functions>
+""")
+    # keywords from all refsect2 and refsect3
+    refsect2 = etree.XPath('//refsect2[@role]')
+    refsect3_enum = etree.XPath('refsect3[@role="enum_members"]/informaltable/tgroup/tbody/row[@role="constant"]')
+    refsect3_enum_details = etree.XPath('entry[@role="enum_member_name"]/para')
+    refsect3_struct = etree.XPath('refsect3[@role="struct_members"]/informaltable/tgroup/tbody/row[@role="member"]')
+    refsect3_struct_details = etree.XPath('entry[@role="struct_member_name"]/para/structfield')
+    for node in files:
+        base_link = node.filename + '#'
+        refsect2_nodes = refsect2(node.xml)
+        for refsect2_node in refsect2_nodes:
+            result.append(create_devhelp2_refsect2_keyword(refsect2_node, base_link))
+            refsect3_nodes = refsect3_enum(refsect2_node)
+            for refsect3_node in refsect3_nodes:
+                details_node = refsect3_enum_details(refsect3_node)[0]
+                name = details_node.attrib['id']
+                result.append(create_devhelp2_refsect3_keyword(refsect3_node, base_link, details_node.text, name))
+            refsect3_nodes = refsect3_struct(refsect2_node)
+            for refsect3_node in refsect3_nodes:
+                details_node = refsect3_struct_details(refsect3_node)[0]
+                name = details_node.attrib['id']
+                result.append(create_devhelp2_refsect3_keyword(refsect3_node, base_link, name, name))
+
+    result.append("""  </functions>
+</book>
+""")
+    return result
+
+
 def create_devhelp2(out_dir, module, xml, files):
     with open(os.path.join(out_dir, module + '.devhelp2'), 'wt',
               newline='\n', encoding='utf-8') as idx:
-        bookinfo_nodes = xml.xpath('/book/bookinfo')
-        title = ''
-        if bookinfo_nodes is not None:
-            bookinfo = bookinfo_nodes[0]
-            title = bookinfo.xpath('./title/text()')[0]
-            online_url = bookinfo.xpath('./releaseinfo/ulink[@role="online-location"]/@url')[0]
-            # TODO: support author too (see devhelp2.xsl)
-        # TODO: fixxref uses '--src-lang' to set the language
-        result = [
-            """<?xml version="1.0" encoding="utf-8" standalone="no"?>
-<book xmlns="http://www.devhelp.net/book" title="%s" link="index.html" author="" name="%s" version="2" language="c" online="%s">
-  <chapters>
-""" % (title, module, online_url)
-        ]
-        # toc
-        result.extend(create_devhelp2_toc(files[0].root))
-        result.append("""  </chapters>
-  <functions>
-""")
-        # keywords from all refsect2 and refsect3
-        refsect2 = etree.XPath('//refsect2[@role]')
-        refsect3_enum = etree.XPath('refsect3[@role="enum_members"]/informaltable/tgroup/tbody/row[@role="constant"]')
-        refsect3_enum_details = etree.XPath('entry[@role="enum_member_name"]/para')
-        refsect3_struct = etree.XPath('refsect3[@role="struct_members"]/informaltable/tgroup/tbody/row[@role="member"]')
-        refsect3_struct_details = etree.XPath('entry[@role="struct_member_name"]/para/structfield')
-        for node in files:
-            base_link = node.filename + '#'
-            refsect2_nodes = refsect2(node.xml)
-            for refsect2_node in refsect2_nodes:
-                result.append(create_devhelp2_refsect2_keyword(refsect2_node, base_link))
-                refsect3_nodes = refsect3_enum(refsect2_node)
-                for refsect3_node in refsect3_nodes:
-                    details_node = refsect3_enum_details(refsect3_node)[0]
-                    name = details_node.attrib['id']
-                    result.append(create_devhelp2_refsect3_keyword(refsect3_node, base_link, details_node.text, name))
-                refsect3_nodes = refsect3_struct(refsect2_node)
-                for refsect3_node in refsect3_nodes:
-                    details_node = refsect3_struct_details(refsect3_node)[0]
-                    name = details_node.attrib['id']
-                    result.append(create_devhelp2_refsect3_keyword(refsect3_node, base_link, name, name))
-
-        result.append("""  </functions>
-</book>
-""")
-        for line in result:
+        for line in create_devhelp2_content(module, xml, files):
             idx.write(line)
 
 
-def get_dirs(uninstalled):
-    if uninstalled:
-        # this does not work from buiddir!=srcdir
-        gtkdocdir = os.path.split(sys.argv[0])[0]
-        if not os.path.exists(gtkdocdir + '/gtk-doc.xsl'):
-            # try 'srcdir' (set from makefiles) too
-            if os.path.exists(os.environ.get("ABS_TOP_SRCDIR", '') + '/gtk-doc.xsl'):
-                gtkdocdir = os.environ['ABS_TOP_SRCDIR']
-        styledir = gtkdocdir + '/style'
-    else:
-        gtkdocdir = os.path.join(config.datadir, 'gtk-doc/data')
-        styledir = gtkdocdir
-    return (gtkdocdir, styledir)
+class FakeDTDResolver(etree.Resolver):
+    """Don't load the docbookx.dtd since we disable the validation anyway.
+
+    libxsml2 does not cache DTDs. If we produce a docbook file with 100 chunks
+    loading such a doc with xincluding will load and parse the docbook DTD 100
+    times. This cases tons of memory allocations and is slow.
+    """
+
+    def resolve(self, url, id, context):
+        if not url.endswith('.dtd'):
+            return None
+        # http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd
+        return self.resolve_string('', context)
 
 
 def main(module, index_file, out_dir, uninstalled, src_lang, paths):
 
     # == Loading phase ==
-    # the next 3 steps could be done in paralel
+    # the next 3 steps could be done in parallel
 
     # 1) load the docuemnt
     _t = timer()
-    # does not seem to be faster
-    # parser = etree.XMLParser(dtd_validation=False, collect_ids=False)
-    # tree = etree.parse(index_file, parser)
-    tree = etree.parse(index_file)
+    parser = etree.XMLParser(dtd_validation=False, collect_ids=False)
+    parser.resolvers.add(FakeDTDResolver())
+    tree = etree.parse(index_file, parser)
     logging.warning("1a: %7.3lf: load doc", timer() - _t)
     _t = timer()
     tree.xinclude()
@@ -1782,20 +1784,18 @@ def main(module, index_file, out_dir, uninstalled, src_lang, paths):
     # 2) copy datafiles
     _t = timer()
     # TODO: handle additional images
-    (gtkdocdir, styledir) = get_dirs(uninstalled)
+    (gtkdocdir, styledir) = config.get_dirs(uninstalled)
     # copy navigation images and stylesheets to html directory ...
     css_file = os.path.join(styledir, 'style.css')
     for f in glob(os.path.join(styledir, '*.png')) + [css_file]:
         shutil.copy(f, out_dir)
-    css_file = os.path.join(out_dir, 'style.css')
-    with open(css_file, 'at', newline='\n', encoding='utf-8') as css:
-        css.write(HTML_FORMATTER.get_style_defs())
+    highlight.append_style_defs(os.path.join(out_dir, 'style.css'))
     logging.warning("2: %7.3lf: copy datafiles", timer() - _t)
 
     # 3) load xref targets
     _t = timer()
     # TODO: migrate options from fixxref
-    # TODO: ideally explicity specify the files we need, this will save us the
+    # TODO: ideally explicitly specify the files we need, this will save us the
     # globbing and we'll load less files.
     fixxref.LoadIndicies(out_dir, '/usr/share/gtk-doc/html', [])
     logging.warning("3: %7.3lf: load xrefs", timer() - _t)
@@ -1862,15 +1862,10 @@ def run(options):
     module = options.args[0]
     document = options.args[1]
 
-    # TODO: rename to 'html' later on
-    # - right now in mkhtml, the dir is created by the Makefile and mkhtml
-    #   outputs into the working directory
-    out_dir = os.path.join(os.path.dirname(document), 'db2html')
-    try:
-        os.mkdir(out_dir)
-    except OSError as e:
-        if e.errno != errno.EEXIST:
-            raise
+    output_dir = options.output_dir or os.getcwd()
+    if options.output_dir and not os.path.isdir(output_dir):
+        os.mkdir(output_dir)
 
-    sys.exit(main(module, document, out_dir, options.uninstalled, options.src_lang,
+    # TODO: pass options.extra_dir
+    sys.exit(main(module, document, output_dir, options.uninstalled, options.src_lang,
                   options.path))
