@@ -59,7 +59,7 @@ MAIN_CODE = """
 #ifdef GTK_IS_WIDGET_CLASS
 #include <gtk/gtk.h>
 #endif
-static GType object_types[$ntypes];
+static GType object_types[${ntypes}];
 
 static GType *
 get_object_types (void)
@@ -95,11 +95,12 @@ ${get_types}
  */
 
 /* The output files */
-const gchar *signals_filename = "$new_signals_filename";
-const gchar *hierarchy_filename = "$new_hierarchy_filename";
-const gchar *interfaces_filename = "$new_interfaces_filename";
-const gchar *prerequisites_filename = "$new_prerequisites_filename";
-const gchar *args_filename = "$new_args_filename";
+const gchar *signals_filename = "${new_signals_filename}";
+const gchar *hierarchy_filename = "${new_hierarchy_filename}";
+const gchar *interfaces_filename = "${new_interfaces_filename}";
+const gchar *prerequisites_filename = "${new_prerequisites_filename}";
+const gchar *args_filename = "${new_args_filename}";
+const gchar *actions_filename = "${new_actions_filename}";
 
 static void output_signals (void);
 static void output_object_signals (FILE *fp,
@@ -125,6 +126,9 @@ static void output_prerequisites (FILE *fp,
 static void output_args (void);
 static void output_object_args (FILE *fp, GType object_type);
 
+static void output_actions (void);
+static void output_object_actions (FILE *fp, GType object_type);
+
 int
 main (${main_func_params})
 {
@@ -137,6 +141,7 @@ main (${main_func_params})
   output_object_interfaces ();
   output_interface_prerequisites ();
   output_args ();
+  output_actions ();
 
   return 0;
 }
@@ -182,7 +187,8 @@ output_object_signals (FILE *fp, GType object_type)
     object_class_name = g_type_name (object_type);
 
     signals = g_signal_list_ids (object_type, &n_signals);
-    qsort (signals, n_signals, sizeof (guint), compare_signals);
+    if (n_signals > 0)
+      qsort (signals, n_signals, sizeof (guint), compare_signals);
 
     for (sig = 0; sig < n_signals; sig++) {
        output_object_signal (fp, object_class_name, signals[sig]);
@@ -333,29 +339,37 @@ get_type_name (GType type, gboolean * is_pointer)
 
   switch (type) {
   case G_TYPE_NONE:
-  case G_TYPE_CHAR:
   case G_TYPE_UCHAR:
   case G_TYPE_BOOLEAN:
-  case G_TYPE_INT:
   case G_TYPE_UINT:
   case G_TYPE_LONG:
   case G_TYPE_ULONG:
-  case G_TYPE_FLOAT:
-  case G_TYPE_DOUBLE:
   case G_TYPE_POINTER:
     /* These all have normal C type names so they are OK. */
     return type_name;
 
+  case G_TYPE_CHAR:
+    return "char";
+
+  case G_TYPE_INT:
+    return "int";
+
+  case G_TYPE_FLOAT:
+    return "float";
+
+  case G_TYPE_DOUBLE:
+    return "double";
+
   case G_TYPE_STRING:
     /* A GtkString is really a gchar*. */
     *is_pointer = TRUE;
-    return "gchar";
+    return "char";
 
   case G_TYPE_ENUM:
   case G_TYPE_FLAGS:
     /* We use a gint for both of these. Hopefully a subtype with a decent
        name will be registered and used instead, as GTK+ does itself. */
-    return "gint";
+    return "int";
 
   case G_TYPE_BOXED:
     /* The boxed type shouldn't be used itself, only subtypes. Though we
@@ -413,7 +427,7 @@ output_object_hierarchy (void)
   FILE *fp;
   gint i,j;
   GType root, type;
-  GType root_types[$ntypes] = { G_TYPE_INVALID, };
+  GType root_types[${ntypes}] = { G_TYPE_INVALID, };
 
   fp = fopen (hierarchy_filename, "w");
   if (fp == NULL) {
@@ -564,6 +578,71 @@ output_prerequisites (FILE  *fp,
 #endif
 }
 
+static void
+output_actions (void)
+{
+  FILE *fp;
+  gint i;
+
+  fp = fopen (actions_filename, "w");
+  if (fp == NULL) {
+    g_warning ("Couldn't open output file: %s : %s", actions_filename, g_strerror(errno));
+    return;
+  }
+
+  for (i = 0; object_types[i]; i++) {
+    output_object_actions (fp, object_types[i]);
+  }
+
+  fclose (fp);
+}
+
+static void
+output_object_actions (FILE *fp, GType object_type)
+{
+  gpointer class;
+
+  if (!G_TYPE_IS_OBJECT (object_type))
+    return;
+
+  class = g_type_class_peek (object_type);
+  if (!class)
+    return;
+
+#ifdef GTK_IS_WIDGET_CLASS
+#if GTK_CHECK_VERSION(3,96,0)
+  if (GTK_IS_WIDGET_CLASS (class)) {
+    guint i = 0;
+    const char *action_name;
+    GType owner;
+    const GVariantType *parameter_type;
+    const char *property_name;
+    const gchar *object_class_name;
+
+    object_class_name = g_type_name (object_type);
+    while (gtk_widget_class_query_action (GTK_WIDGET_CLASS (class),
+                                          i,
+                                          &owner,
+                                          &action_name,
+                                          &parameter_type,
+                                          &property_name)) {
+      i++;
+      if (owner == G_TYPE_FROM_CLASS (class))
+        fprintf (fp, "<ACTION>\\n"
+                     "<NAME>%s:::%s</NAME>\\n"
+                     "<PARAMETER>%s</PARAMETER>\\n"
+                     "<PROPERTY>%s</PROPERTY>\\n"
+                     "</ACTION>\\n\\n",
+                 object_class_name,
+                 action_name,
+                 parameter_type ? g_variant_type_peek_string (parameter_type) : "",
+                 property_name ? property_name : "");
+    }
+  }
+#endif
+#endif
+}
+
 static void
 output_args (void)
 {
@@ -611,11 +690,11 @@ describe_double_constant (gdouble value)
     desc = g_strdup ("G_MINDOUBLE");
   else if (GTKDOC_COMPARE_FLOAT (value, -G_MAXDOUBLE))
     desc = g_strdup ("-G_MAXDOUBLE");
-  else if (GTKDOC_COMPARE_FLOAT (value, G_MAXFLOAT))
+  else if (GTKDOC_COMPARE_FLOAT (value, (gdouble)G_MAXFLOAT))
     desc = g_strdup ("G_MAXFLOAT");
-  else if (GTKDOC_COMPARE_FLOAT (value, G_MINFLOAT))
+  else if (GTKDOC_COMPARE_FLOAT (value, (gdouble)G_MINFLOAT))
     desc = g_strdup ("G_MINFLOAT");
-  else if (GTKDOC_COMPARE_FLOAT (value, -G_MAXFLOAT))
+  else if (GTKDOC_COMPARE_FLOAT (value, (gdouble)-G_MAXFLOAT))
     desc = g_strdup ("-G_MAXFLOAT");
   else{
     /* make sure floats are output with a decimal dot irrespective of
@@ -690,10 +769,12 @@ describe_unsigned_constant (gsize size, guint64 value)
         else if (value == G_MAXUINT)
           desc = g_strdup ("G_MAXUINT");
       }
-      if (value == (guint64)G_MAXLONG)
-        desc = g_strdup ("G_MAXLONG");
-      else if (value == G_MAXULONG)
-        desc = g_strdup ("G_MAXULONG");
+      if (desc == NULL) {
+        if (value == (guint64)G_MAXLONG)
+          desc = g_strdup ("G_MAXLONG");
+        else if (value == G_MAXULONG)
+          desc = g_strdup ("G_MAXULONG");
+      }
       break;
     case 8:
       if (value == G_MAXINT64)
@@ -1093,7 +1174,8 @@ output_object_args (FILE *fp, GType object_type)
   style_prop = FALSE;
 
   while (TRUE) {
-    qsort (properties, n_properties, sizeof (GParamSpec *), compare_param_specs);
+    if (n_properties > 0)
+      qsort (properties, n_properties, sizeof (GParamSpec *), compare_param_specs);
     for (arg = 0; arg < n_properties; arg++) {
       GParamSpec *spec = properties[arg];
       const gchar *nick, *blurb, *dot;
@@ -1149,12 +1231,14 @@ output_object_args (FILE *fp, GType object_type)
     g_free (properties);
 
 #ifdef GTK_IS_CONTAINER_CLASS
+#if !GTK_CHECK_VERSION(3,96,0)
     if (!child_prop && GTK_IS_CONTAINER_CLASS (class)) {
       properties = gtk_container_class_list_child_properties (class, &n_properties);
       child_prop = TRUE;
       continue;
     }
 #endif
+#endif
 
 #ifdef GTK_IS_CELL_AREA_CLASS
     if (!child_prop && GTK_IS_CELL_AREA_CLASS (class)) {
@@ -1183,14 +1267,14 @@ MAIN_CODE_END = """
 """
 
 
-def execute_command(options, description, command):
+def execute_command(options, description, command, env=None):
     if options.verbose:
         call = subprocess.check_call
     else:
         call = subprocess.check_output
 
     try:
-        call(command)
+        call(command, env=env)
     except subprocess.CalledProcessError as e:
         logging.warning('%s scanner failed: %d, command: %s', description,
                         e.returncode, ' '.join(command))
@@ -1203,8 +1287,10 @@ def execute_command(options, description, command):
 
 
 def run(options):
+    logging.info('options: %s', str(options.__dict__))
+
     c_file = options.module + '-scan.c'
-    output = common.open_text(c_file, 'w')
+    output = open(c_file, 'w', encoding='utf-8')
 
     base_filename = os.path.join(options.output_dir, options.module)
     old_signals_filename = base_filename + '.signals'
@@ -1217,6 +1303,8 @@ def run(options):
     new_prerequisites_filename = base_filename + '.prerequisites.new'
     old_args_filename = base_filename + '.args'
     new_args_filename = base_filename + '.args.new'
+    old_actions_filename = base_filename + '.actions'
+    new_actions_filename = base_filename + '.actions.new'
 
     # generate a C program to scan the types
 
@@ -1225,7 +1313,7 @@ def run(options):
     get_types = ""
     ntypes = 1
 
-    for line in common.open_text(options.types):
+    for line in open(options.types, 'r', encoding='utf-8'):
         if line.startswith('#include'):
             includes += line
         elif line.startswith('%') or line.strip() == '':
@@ -1256,6 +1344,19 @@ def run(options):
     elif "argc" not in type_init_func and "argv" not in type_init_func:
         main_func_params = "void"
 
+    # TODO: we should explicitly reference the vars that we substitute
+    # template_vars = {
+    #     'get_types': get_types,
+    #     'ntypes': ntypes,
+    #     'main_func_params', main_func_params,
+    #     'new_args_filename': new_args_filename,
+    #     'new_hierarchy_filename': new_hierarchy_filename,
+    #     'new_interfaces_filename': new_interfaces_filename,
+    #     'new_prerequisites_filename': new_prerequisites_filename,
+    #     'new_signals_filename': new_signals_filename,
+    #     'type_init_func': type_init_func
+    # }
+    # output.write(string.Template(MAIN_CODE).substitute(template_vars))
     output.write(string.Template(MAIN_CODE).substitute(locals()))
 
     if options.query_child_properties:
@@ -1287,8 +1388,12 @@ def run(options):
     if res:
         return res
 
+    run_env = os.environ.copy()
+    run_env['LC_MESSAGES'] = 'C'
+    run_env.pop('LC_ALL', None)
     res = execute_command(options, 'Running',
-                          shlex.split(options.run) + ['./' + x_file])
+                          shlex.split(options.run) + ['./' + x_file],
+                          env=run_env)
     if res:
         return res
 
@@ -1310,5 +1415,6 @@ def run(options):
     common.UpdateFileIfChanged(old_interfaces_filename, new_interfaces_filename, False)
     common.UpdateFileIfChanged(old_prerequisites_filename, new_prerequisites_filename, False)
     common.UpdateFileIfChanged(old_args_filename, new_args_filename, False)
+    common.UpdateFileIfChanged(old_actions_filename, new_actions_filename, False)
 
     return 0
